@@ -13,7 +13,7 @@
 
 // You should have received a copy of the GNU General Public License along
 // with this library; see the file COPYING.  If not, write to the Free
-// Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+// Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
 // USA.
 
 // 27.6.2.7 standard basic_ostream manipulators
@@ -26,7 +26,6 @@ void test01(void)
 {
   bool test __attribute__((unused)) = true;
 
-  const wchar_t str_lit01[] = L"  venice ";
   const std::wstring str01(L" santa barbara ");
   std::wstring str04;
   std::wstring str05;
